# -*- coding: utf-8 -*-
"""
Created on Tue Nov 12 13:49:24 2019

@author: Timon Schapeler (timons@mail.upb.de)

(based on Lundeen/Feito MATLAB code from http://www.photonicquantum.info/Tools.html)
"""
### Imports ###
import cvxpy as cp # <-- has to be installed beforehand
import numpy as np
import matplotlib.pyplot as plt

### Load in the data used for the reconstruction ###
# For the example use the "prob_dist-Template.py" file and generate the matrices F and P
imported_F = np.load('F_example.npy')
imported_P = np.load('P_example.npy')

### Defining functions ###
def reconstruction(imported_F, imported_P, gamma):
    """
    This function will reconstruct the POVM elements based on the given F and P matrices.

    Parameters
    ----------
    F : np.array, DxM matrix
        Matrix containing the coherent state (expanded in Fock basis 
        up to cutoff value M calculated with the function 'calc_M').
    P : np.array, DxN matrix
        Matrix containing the probability distribution of the detector..
    gamma : float, between 0 and 1
        Smoothing factor.

    Returns
    -------
    reconstructed_POVMs : np.array, MxN matrix
        Matrix containing the reconstructed POVM elements of the detector.

    """
    # Get dimensions of the matricies
    N = imported_P.shape[1] #number of possible outcomes of the detector
    D = imported_P.shape[0] #number of probe states
    M = imported_F.shape[1] #fock space cuttoff / hilbert space dimension
    
    # Setup the variable
    rec_POVM = cp.Variable((M,N), nonneg=True) #nonneg makes sure, that all elements of the POVM are positive, this is one of the two constraints
    
    # Setup smoothing
    smooth = 0
    for k in range(0,N):
        for i in range(0,M-1):
            smooth = smooth + (rec_POVM[i,k] - rec_POVM[i+1,k])**2
    
    # Add the second constraint (summation to one)
    constraints_list = []
    for i in range(M):
        constraints_list.append(sum(rec_POVM[i,:]) == 1) #this adds the constraint that every row of the POVMs should sum to one (not strictly, due to numerics)
    
    # Define objective
    obj = cp.Minimize(cp.norm(imported_P - imported_F * rec_POVM, 'fro') + gamma*smooth) #define the optimization problem
    problem = cp.Problem(obj, constraints = constraints_list)
    
    # Solve the problem
    problem.solve(solver=cp.SCS, verbose=True)  #Returns the optimal value. Default solver: cp.SCS (others: CVXOPT, ECOS)
    print("status:", problem.status)
    print("optimal value", problem.value)
    #print("optimal var", rec_POVM.value)
    
    reconstructed_POVMs = np.asarray(rec_POVM.value) #convert the output of the optimization to a np.array
    np.save('rec_POVM_gamma' + str(gamma) + '.npy', reconstructed_POVMs)      

    return reconstructed_POVMs

def plotting(reconstructed_POVMs, gamma, m):
    """
    Parameters
    ----------
    reconstructed_POVMs: np.array, MxN matrix
        Matrix containing the reconstructed POVM elements of the detector.
    gamma : float, has to match the one used for the reconstruction
        Which gamma value was used in the reconstruction? (Here used for naming the plot.)    
    m : int
        Photon number in Fock space (|m><m|) up to which to plot the POVMs. Cannot exceed M.

    Returns
    -------
    None.

    """
    N = reconstructed_POVMs.shape[1] #number of possible outcomes of the detector
    
    # Plot the POVMs by plotting each column of the reconstructed POVM vs. photon number
    plt.figure(1, figsize=(16,8))
    for i in range(N):
        plt.subplot(2,3,i+1)
        plt.bar(np.arange(m), reconstructed_POVMs[:m,i], width=0.8, label='n='+str(i))
        plt.ylim(0,1)
        plt.legend(frameon=True, handlelength=0, handletextpad=-0.0, loc='best')
        plt.xlabel('Photon number, i')
        plt.ylabel(r'$\theta_i^{(n)}$')
    plt.tight_layout()
    pltname = 'plot_gamma' + str(gamma) + 'm' + str(m) + '.png'
    plt.savefig(pltname, dpi=300)
    plt.show()


POVM = reconstruction(imported_F, imported_P, 0.1) #reconsructs the POVM elements for the example F and P matries and a smoothing parameter of 0.1

plotting(POVM, 0.1, 100) #creates a simple bar plot of the POVMs to visualize them 

