# -*- coding: utf-8 -*-
"""
Created on Wed Nov  6 14:15:54 2019

@author: Timon Schapeler
"""

### Imports ###
import numpy as np
import scipy.special
from scipy.stats import poisson

### Defining functions ###
def calc_M(max_amp):
    """
    This function will calculate the cutoff photon number M (in Fock basis) 
    for the maximal coherent amplitude used in the experiment.    

    Parameters
    ----------
    max_amp : float
        Largest mean photon number.

    Returns
    -------
    cutoff_top : float
        Value of M, where the Poissonian distribution can be cut off savely
        without losing too much information (mean photon number + 6*sigma).
    cutoff_bottom : float
        (mean photon number - 6*sigma)
    """

    std = scipy.stats.poisson.std(max_amp) # np.sqrt(max_amp) would do the same
    cutoff_bottom = int(np.floor(max_amp - 6*std))
    cutoff_top = int(np.ceil(max_amp + 6*std)) # calculate cutoff, which is: mean photon number + 6*sigma, then round up and take integer value
    # pdist = poisson.pmf(np.arange(0,1.5*cutoff_top), max_amp) # OPTIONAL: calculates the distribution 

    return cutoff_bottom, cutoff_top

def calc_F(coherent_amplitude_list):
    """
    This function will calculate the matrix F containing all the coherent probe states.

    Parameters
    ----------
    coherent_amplitude_list : np.array
        List of mean photon numbers that determine the Poisson distributions.

    Returns
    -------
    F : np.array, DxM matrix
        Matrix containing the coherent state (expanded in Fock basis 
        up to cutoff value M calculated with the function 'calc_M').
    """
    D = len(coherent_amplitude_list) #total number of input states
    M = calc_M(max(coherent_amplitude_list))[1] #cutoff value for the photon number distribution
    F = np.zeros((D,M)) #input state matrix
    for k,l in enumerate(coherent_amplitude_list):
        F[k,:] = poisson.pmf(np.arange(0,M), l)
        
    return F

def calc_P(N, F, eta):
    '''
    This function generates a theoretical probability distribution for an artificial detector with N outcomes and an efficiency of eta. 

    Parameters
    ----------
    N : float
        Number of outcomes of the detector.
    F : np.array, DxM matrix (calculated with 'calc_F'-function)
        Matrix containing the coherent state.
    eta : float, between 0 and 1
        Relates to loss, eta=1: no loss, perfect detector.

    Returns
    -------
    P : np.array, DxN matrix
        Matrix containing the probability distribution of the detector.
    '''
    # Create a diagonal POVM, that can detect from 0 to N-1 clicks but
    # cannot distinguish "N-1" from "N-1 or more" photons
    M = F.shape[1]
    global POVM
    POVM = np.eye(M,N)
    POVM[N-1:M,N-1] = 1
    
    # Include loss (loss matrix L)
    global L
    L = np.zeros((M,M))
    for i in range(0,M):
        for j in range(0,i+1):
            L[j,i] = scipy.special.binom(i,j)*eta**j*(1-eta)**(i-j)
    global lossy_POVM
    lossy_POVM = np.matmul(L.transpose(),POVM)
    for i in range(len(lossy_POVM[:,0])):
        if lossy_POVM[i,N-1] == 1:
            lossy_POVM[i:,:N-1] = 0
            lossy_POVM[i:,N-1] = 1 
   
    # Generate the Probability Distribution  
    P = np.matmul(F,lossy_POVM)#.transpose()
    return P

### Using the functions to generate the F and P matrices ###
F = calc_F(np.arange(0,11)*10) #generates coherent input state matrix with mean photon numbers of 0 to 100 in steps of 10
P = calc_P(5, F, 0.1) #generates outcome matrix for N=5 outcomes, the previously generated F matrix and and efficiency of 0.1

np.save('F_example.npy',F) #save the matrices
np.save('P_example.npy',P)
